//==================================================================================================
// CONCEPTS FOR TEMPLATE SPECIALIZATION IN C++                         examples/concept-based_2.cpp
//                                                      By Bruno Bachelet, Antoine Mahul & Loic Yon
//==================================================================================================
// Copyright (c) 2010-2016
// Bruno Bachelet - bruno@nawouak.net - http://www.nawouak.net
// Antoine Mahul  - antoine.mahul@clermont-universite.fr
// Loic Yon       - loic.yon@isima.fr - http://www.isima.fr/~loic
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the license, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details (http://www.gnu.org).
//
// This file shows a C++03 compliant example of static specialization with the C4TS++ library (*).
//
// (*) Bruno Bachelet, Antoine Mahul and Loic Yon.
//     Template Metaprogramming Techniques for Concept-Based Specialization.
//     Research report LIMOS/RR10-18, LIMOS, UMR 6158-CNRS,
//     Universite Blaise Pascal, BP 10125, 63173 Aubiere, France.

// Headers //---------------------------------------------------------------------------------------
#include <iostream>
#include <typeinfo>
#include <gnx_concept.hpp>

// Concepts Taxonomy //-----------------------------------------------------------------------------

//------------------------------------------------------------------------------Concepts Declaration
gnx_declare_concept(Concept1);
gnx_declare_concept(Concept2);
gnx_declare_concept(Concept3);
gnx_declare_concept(Concept4);
gnx_declare_concept(Concept5);
gnx_declare_concept(Concept6);

//-------------------------------------------------------------------------------Classes Declaration
class ClassA {};
class ClassB {};
class ClassC {};
class ClassD {};

//---------------------------------------------------------------Modeling & Refinement Relationships
template <> struct gnx_models_concept<ClassA,Concept3> : gnx_true {};
template <> struct gnx_models_concept<ClassA,Concept4> : gnx_true {};
template <> struct gnx_models_concept<ClassB,Concept4> : gnx_true {};
template <> struct gnx_models_concept<ClassC,Concept4> : gnx_true {};
template <> struct gnx_models_concept<ClassC,Concept5> : gnx_true {};
template <> struct gnx_models_concept<ClassD,Concept5> : gnx_true {};
template <> struct gnx_models_concept<ClassD,Concept6> : gnx_true {};

template <> struct gnx_models_concept<Concept2,Concept1> : gnx_true {};
template <> struct gnx_models_concept<Concept6,Concept1> : gnx_true {};
template <> struct gnx_models_concept<Concept3,Concept2> : gnx_true {};

// Trait Class //-----------------------------------------------------------------------------------
struct TraitContext;

template <class TYPE,class = gnx_best_concept(TraitContext,TYPE)>
struct Trait : gnx_value<int,0> {};

//----------------------------------------------------------------------------------Concept1 Version
template <> struct gnx_uses_concept<TraitContext,Concept1> : gnx_true {};

template <class TYPE> struct Trait<TYPE,Concept1> : gnx_value<int,1> {};

//----------------------------------------------------------------------------------Concept2 Version
template <> struct gnx_uses_concept<TraitContext,Concept2> : gnx_true {};

template <class TYPE> struct Trait<TYPE,Concept2> : gnx_value<int,2> {};

//----------------------------------------------------------------------------------Concept5 Version
template <> struct gnx_uses_concept<TraitContext,Concept5> : gnx_true {};

template <class TYPE> struct Trait<TYPE,Concept5> : gnx_value<int,5> {};

//----------------------------------------------------------------------------------Concept7 Version
/* Uncomment the following lines to test the "gnx_contextual_concept" instantiation problem.
   "Concept7" and its associated relationships will not be considered. */

/*
// To force "gnx_contextual_concept" instantiation. //
template <> struct Trait<int> : gnx_value<int,-1> {};

// New concept //
gnx_declare_concept(Concept7);

// New relationship //
template <> struct gnx_models_concept<Concept4,Concept7> : gnx_true {};

// New specialization //
template <> struct gnx_uses_concept<TraitContext,Concept7> : gnx_true {};

template <class TYPE> struct Trait<TYPE,Concept7> : gnx_value<int,7> {};
*/

//-------------------------------------------------------------------Concept1 vs. Concept5 Ambiguity
/* The following line forces "Concept1" to be the best concept for "ClassD",
   because "Concept1" and "Concept5" are equal candidates. */

// template <> struct gnx_contextual_concept<TraitContext,ClassD> : gnx_type<Concept1> {};

// Test //------------------------------------------------------------------------------------------
int main(void) {
 std::cout << "A matches 1 ? " << gnx_matches_concept<ClassA,Concept1>::value << std::endl;
 std::cout << "A matches 2 ? " << gnx_matches_concept<ClassA,Concept2>::value << std::endl;
 std::cout << "A matches 3 ? " << gnx_matches_concept<ClassA,Concept3>::value << std::endl;
 std::cout << "A matches 4 ? " << gnx_matches_concept<ClassA,Concept4>::value << std::endl;
 std::cout << "A matches 5 ? " << gnx_matches_concept<ClassA,Concept5>::value << std::endl;
 std::cout << "A matches 6 ? " << gnx_matches_concept<ClassA,Concept6>::value << std::endl;
 std::cout << std::endl;

 std::cout << "B matches 1 ? " << gnx_matches_concept<ClassB,Concept1>::value << std::endl;
 std::cout << "B matches 2 ? " << gnx_matches_concept<ClassB,Concept2>::value << std::endl;
 std::cout << "B matches 3 ? " << gnx_matches_concept<ClassB,Concept3>::value << std::endl;
 std::cout << "B matches 4 ? " << gnx_matches_concept<ClassB,Concept4>::value << std::endl;
 std::cout << "B matches 5 ? " << gnx_matches_concept<ClassB,Concept5>::value << std::endl;
 std::cout << "B matches 6 ? " << gnx_matches_concept<ClassB,Concept6>::value << std::endl;
 std::cout << std::endl;

 std::cout << "C matches 1 ? " << gnx_matches_concept<ClassC,Concept1>::value << std::endl;
 std::cout << "C matches 2 ? " << gnx_matches_concept<ClassC,Concept2>::value << std::endl;
 std::cout << "C matches 3 ? " << gnx_matches_concept<ClassC,Concept3>::value << std::endl;
 std::cout << "C matches 4 ? " << gnx_matches_concept<ClassC,Concept4>::value << std::endl;
 std::cout << "C matches 5 ? " << gnx_matches_concept<ClassC,Concept5>::value << std::endl;
 std::cout << "C matches 6 ? " << gnx_matches_concept<ClassC,Concept6>::value << std::endl;
 std::cout << std::endl;

 std::cout << "D matches 1 ? " << gnx_matches_concept<ClassD,Concept1>::value << std::endl;
 std::cout << "D matches 2 ? " << gnx_matches_concept<ClassD,Concept2>::value << std::endl;
 std::cout << "D matches 3 ? " << gnx_matches_concept<ClassD,Concept3>::value << std::endl;
 std::cout << "D matches 4 ? " << gnx_matches_concept<ClassD,Concept4>::value << std::endl;
 std::cout << "D matches 5 ? " << gnx_matches_concept<ClassD,Concept5>::value << std::endl;
 std::cout << "D matches 6 ? " << gnx_matches_concept<ClassD,Concept6>::value << std::endl;
 std::cout << std::endl;

 typedef gnx_contextual_concept<TraitContext,ClassA>::type ConceptA;
 typedef gnx_contextual_concept<TraitContext,ClassB>::type ConceptB;
 typedef gnx_contextual_concept<TraitContext,ClassC>::type ConceptC;
 typedef gnx_contextual_concept<TraitContext,ClassD>::type ConceptD;

 std::cout << "best_concept<A> = " << typeid(ConceptA).name() << std::endl;
 std::cout << "best_concept<B> = " << typeid(ConceptB).name() << std::endl;
 std::cout << "best_concept<C> = " << typeid(ConceptC).name() << std::endl;
 std::cout << "best_concept<D> = " << typeid(ConceptD).name() << std::endl;
 std::cout << std::endl;

 std::cout << "Trait<A> = " << Trait<ClassA>::value << std::endl;
 std::cout << "Trait<B> = " << Trait<ClassB>::value << std::endl;
 std::cout << "Trait<C> = " << Trait<ClassC>::value << std::endl;
 std::cout << "Trait<D> = " << Trait<ClassD>::value << std::endl;

 return 0;
}

// End //-------------------------------------------------------------------------------------------
