//==================================================================================================
// CONCEPTS FOR TEMPLATE SPECIALIZATION IN C++                                 examples/testing.hpp
//                                                      By Bruno Bachelet, Antoine Mahul & Loic Yon
//==================================================================================================
// Copyright (c) 2010-2016
// Bruno Bachelet - bruno@nawouak.net - http://www.nawouak.net
// Antoine Mahul  - antoine.mahul@clermont-universite.fr
// Loic Yon       - loic.yon@isima.fr - http://www.isima.fr/~loic
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the license, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details (http://www.gnu.org).
//
// This file contains the testing procedure used in the different examples.

// Guardian //--------------------------------------------------------------------------------------
#ifndef _C4TS_EXAMPLES_TESTING_HPP_
#define _C4TS_EXAMPLES_TESTING_HPP_

// Headers //---------------------------------------------------------------------------------------
#include <cstdlib>
#include <iostream>
#include <list>
#include <map>
#include <set>
#include <vector>

// Testing //---------------------------------------------------------------------------------------

//---------------------------------------------------------------------------Operator << (Vector<T>)
template <class T>
std::ostream & operator << (std::ostream & stream, const std::vector<T> & container) {
 for (typename std::vector<T>::const_iterator it = container.begin(); it!=container.end(); ++it)
  stream << *it << " ";

 return stream;
}

//-----------------------------------------------------------------------------Operator << (List<T>)
template <class T>
std::ostream & operator << (std::ostream & stream, const std::list<T> & container) {
 for (typename std::list<T>::const_iterator it = container.begin(); it!=container.end(); ++it)
  stream << *it << " ";

 return stream;
}

//------------------------------------------------------------------------------Operator << (Set<T>)
template <class T>
std::ostream & operator << (std::ostream & stream, const std::set<T> & container) {
 for (typename std::set<T>::const_iterator it = container.begin(); it!=container.end(); ++it)
  stream << *it << " ";

 return stream;
}

//----------------------------------------------------------------------------Operator << (Map<K,T>)
template <class K,class T>
std::ostream & operator << (std::ostream & stream, const std::map<K,T> & container) {
 for (typename std::map<K,T>::const_iterator it = container.begin(); it!=container.end(); ++it)
  stream << it->first << ";" << it->second << " ";

 return stream;
}

//------------------------------------------------------------------------------------TestSerializer
template <class T> void testSerializer(const T & object) {
 char buffer[256];
 T restored = T();
 int nb;

 std::cout << "    Original = " << object << std::endl;
 std::cout << "    Restored = " << restored << std::endl;

 nb=Serializer<T>::deflate(buffer,object,true);
 std::cout << "    Deflate " << nb << " Bytes" << std::endl;

 nb=Serializer<T>::inflate(restored,buffer,true);
 std::cout << "    Inflate " << nb << " Bytes" << std::endl;

 std::cout << "    Restored = " << restored << std::endl;
}

//----------------------------------------------------------------------------------------------Main
int main(void) {
 std::srand(27);

 {
  std::cout << std::endl;
  std::cout << "[>] Serializer<int>" << std::endl;
  testSerializer(std::rand());
 }

 {
  std::vector<int> vector;

  for (int i=0; i<10; ++i) vector.push_back(std::rand());

  std::cout << std::endl;
  std::cout << "[>] Serializer< std::vector<int> >" << std::endl;
  testSerializer(vector);
 }

 {
  std::list<int> list;

  for (int i=0; i<10; ++i) list.push_back(std::rand());

  std::cout << std::endl;
  std::cout << "[>] Serializer< std::list<int> >" << std::endl;
  testSerializer(list);
 }

 /* In example 1, the following code causes ambiguities in the specialization of "Serializer",
    with Embarcadero (Borland) C++ 6.20 and G++ 3.4.5 compilers. */

 {
  std::set<int> set;

  for (int i=0; i<10; ++i) set.insert(std::rand());

  std::cout << std::endl;
  std::cout << "[>] Serializer< std::set<int> >" << std::endl;
  testSerializer(set);
 }

 /* Try to uncomment the following code.
    In example 1, "std::string" matches the type pattern of "std::set".
    In the other examples, "std::string" doesn't match any specialized version of "Serializer",
    and the compiler provides a clear message. */

 /*
 {
  std::string string = "Not working !!!";

  std::cout << std::endl;
  std::cout << "[>] Serializer<std::string>" << std::endl;
  testSerializer(string);
 }
 */

 /* In example 1, the following code causes ambiguities in the specialization of "Serializer",
    with Embarcadero (Borland) C++ 6.20 and G++ 3.4.5 compilers. */

 {
  std::map<int,float> map;

  for (int i=0; i<5; ++i) map.insert(std::make_pair(std::rand(),float(std::rand())));

  std::cout << std::endl;
  std::cout << "[>] Serializer< std::map<int,float> >" << std::endl;
  testSerializer(map);
 }
}

// End //-------------------------------------------------------------------------------------------
#endif
