//==================================================================================================
// CONCEPTS FOR TEMPLATE SPECIALIZATION IN C++                           examples/type_patterns.cpp
//                                                      By Bruno Bachelet, Antoine Mahul & Loic Yon
//==================================================================================================
// Copyright (c) 2010-2016
// Bruno Bachelet - bruno@nawouak.net - http://www.nawouak.net
// Antoine Mahul  - antoine.mahul@clermont-universite.fr
// Loic Yon       - loic.yon@isima.fr - http://www.isima.fr/~loic
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the license, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details (http://www.gnu.org).
//
// This file shows an example of static specialization based on type patterns.

// Headers //---------------------------------------------------------------------------------------
#include <list>
#include <map>
#include <set>
#include <vector>

// Serialization Code //----------------------------------------------------------------------------
#define C4TS_EXAMPLES_SERIALIZATION_NEED_PROTOTYPES
#include <serialization.hpp>

// Serializer Class //------------------------------------------------------------------------------

//-----------------------------------------------------------------------------------Primary Version
template <class T> class Serializer {
 public:
  static int deflate(char * copy,const T & object,bool display = false)
  { return deflate_single_object(copy,object,display); }

  static int inflate(T & object,const char * copy,bool display = false)
  { return inflate_single_object(object,copy,display); }
};

//----------------------------------------------------------------------------------STL Pair Version
template <class T1,class T2> class Serializer< std::pair<T1,T2> > {
 public:
  static int deflate(char * copy,const std::pair<T1,T2> & object,bool display = false)
  { return deflate_stl_pair(copy,object,display); }

  static int inflate(std::pair<T1,T2> & object,const char * copy,bool display = false)
  { return inflate_stl_pair(object,copy,display); }
};

//------------------------------------------------------------------------------STL Sequence Version
template <class T,class ALLOC,template <class,class> class CONTAINER>
class Serializer< CONTAINER<T,ALLOC> > {
 public:
  static int deflate(char * copy,const CONTAINER<T,ALLOC> & object,bool display = false)
  { return deflate_stl_sequence(copy,object,display); }

  static int inflate(CONTAINER<T,ALLOC> & object,const char * copy,bool display = false)
  { return inflate_stl_sequence(object,copy,display); }
};

//-----------------------------------------------------------------------------------STL Set Version
template <class T,class COMP,class ALLOC,
          template <class,class,class> class CONTAINER>
class Serializer< CONTAINER<T,COMP,ALLOC> > {
 public:
  static int deflate(char * copy,const CONTAINER<T,COMP,ALLOC> & object,bool display = false)
  { return deflate_stl_associative(copy,object,display); }

  static int inflate(CONTAINER<T,COMP,ALLOC> & object,const char * copy,bool display = false)
  { return inflate_stl_associative(object,copy,display); }
};

//-----------------------------------------------------------------------------------STL Map Version
template <class K,class T,class COMP,class ALLOC,
          template <class,class,class,class> class CONTAINER>
class Serializer< CONTAINER<K,T,COMP,ALLOC> > {
 public:
  static int deflate(char * copy,const CONTAINER<K,T,COMP,ALLOC> & object,bool display = false)
  { return deflate_stl_associative(copy,object,display); }

  static int inflate(CONTAINER<K,T,COMP,ALLOC> & object,const char * copy,bool display = false)
  { return inflate_stl_associative(object,copy,display); }
};

// Serialization Code //----------------------------------------------------------------------------
#include <serialization.hpp>

// Testing Code //----------------------------------------------------------------------------------
#include <testing.hpp>

// End //-------------------------------------------------------------------------------------------
