//==================================================================================================
// CONCEPTS FOR TEMPLATE SPECIALIZATION IN C++                    experiments/generate_instance.cpp
//                                                      By Bruno Bachelet, Antoine Mahul & Loic Yon
//==================================================================================================
// Copyright (c) 2010-2016
// Bruno Bachelet - bruno@nawouak.net - http://www.nawouak.net
// Antoine Mahul  - antoine.mahul@clermont-universite.fr
// Loic Yon       - loic.yon@isima.fr - http://www.isima.fr/~loic
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the license, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details (http://www.gnu.org).
//
// This file generates a random instance of concept-based specialization for the C4TS++ library.

// Headers //---------------------------------------------------------------------------------------
#include <cmath>
#include <cstdlib>
#include <ctime>
#include <iostream>
#include <set>

using namespace std;

// Instance Generation //---------------------------------------------------------------------------
void generate_instance(unsigned n,unsigned r,unsigned p,char meta) {
 // Header //
 cout << "#include <gnx_concept.hpp>" << endl;
 cout << endl;

 // Concepts Declaration //
 cout << "// n = " << n << endl;
 cout << endl;

 for (unsigned i = 1; i<=n; ++i)
  cout << "gnx_declare_concept(Concept" << i << ");" << endl;

 cout << "gnx_declare_concept(DummyConcept);" << endl;
 cout << endl;

 // Specialization Context //
 cout << "struct DummyContext;" << endl;
 cout << endl;

 for (unsigned i = 1; i<=n; ++i)
  cout << "template <> struct gnx_uses_concept<DummyContext,Concept" << i << "> "
       << ": gnx_true {};" << endl;

 cout << endl;

 // Relationships Declaration //
 srand(time(0));

 typedef set< pair<unsigned,unsigned> > relations_t;

 relations_t relations;

 while (relations.size()<r) {
  unsigned i = rand()%(n-1)+1;
  unsigned j = i+1+rand()%(n-i);

  relations.insert(make_pair(i,j));
 }

 cout << "// r = " << relations.size() << endl;
 cout << endl;

 for (relations_t::const_iterator i = relations.begin(); i!=relations.end(); ++i)
  cout << "template <> struct gnx_models_concept<Concept" << i->first
       << ",Concept" << i->second << "> : gnx_true {};" << endl;

 cout << endl;

 // Metafunction Calls //
 cout << "// p = " << p << endl;
 cout << endl;

 switch (meta) {
  case 'b':
   for (unsigned i=1; i<=p; ++i)
    cout << "typedef gnx_direct_concepts<Concept" << i << ">::type type_" << i << ";" << endl;

   break;

  case 'c':
   for (unsigned i=1; i<=p; ++i)
    cout << "typedef gnx_all_concepts<Concept" << i << ">::type type_" << i << ";" << endl;

   break;

  case 'd':
   for (unsigned i=1; i<=p; ++i)
    cout << "typedef gnx_value<bool,gnx_matches_concept<Concept" << i
              << ",DummyConcept>::value> type_" << i << ";" << endl;

   break;

  case 'e':
   for (unsigned i=1; i<=p; ++i)
    cout << "typedef gnx_contextual_concept<DummyContext,Concept" << i
              << ">::type type_" << i << ";" << endl;

   break;
 }

 // Main //
 cout << endl;
 cout << "int main(void) {" << endl;
 cout << " return 0;" << endl;
 cout << "}" << endl;
 cout << endl;
}

// Main //------------------------------------------------------------------------------------------
int main(int argc,const char * argv[]) {
 if (argc!=5) {
  cout << "Syntax: " << argv[0] << " <n> <r> <p> <x>" << endl;
  cout << "Generates a random instance of concept-based specialization for the C4TS++ library.";
  cout << endl;
  cout << "- n = number of concepts" << endl;
  cout << "- r = number of relationships" << endl;
  cout << "- p = number of calls to the metafunction to be tested" << endl;
  cout << "- x = metafunction to be tested" << endl;
  cout << "      - a = no metafunction" << endl;
  cout << "      - b = 'gnx_direct_concepts' metafunction" << endl;
  cout << "      - c = 'gnx_all_concepts' metafunction" << endl;
  cout << "      - d = 'gnx_matches_concept' metafunction" << endl;
  cout << "      - e = 'gnx_contextual_concept' metafunction" << endl;
  return -1;
 }

 generate_instance(atoi(argv[1]),atoi(argv[2]),atoi(argv[3]),argv[4][0]);

 return 0;
}

// End //-------------------------------------------------------------------------------------------
