//==================================================================================================
// CONCEPTS FOR TEMPLATE SPECIALIZATION IN C++                     c4ts++_optimized/gnx_concept.hpp
//                                                      By Bruno Bachelet, Antoine Mahul & Loic Yon
//==================================================================================================
// Copyright (c) 2010-2016
// Bruno Bachelet - bruno@nawouak.net - http://www.nawouak.net
// Antoine Mahul  - antoine.mahul@clermont-universite.fr
// Loic Yon       - loic.yon@isima.fr - http://www.isima.fr/~loic
//
// This file is part of the C4TS++ library. This library is free software; you can redistribute it
// and/or modify it under the terms of the GNU Library General Public License as published by the
// Free Software Foundation; either version 2 of the License, or (at your option) any later
// version.
//
// This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
// the GNU Library General Public License for more details (http://www.gnu.org).
//
// This file provides metaprogramming elements for concept-based specialization.

// Guardian //--------------------------------------------------------------------------------------
#ifndef _C4TS_GNX_CONCEPT_
#define _C4TS_GNX_CONCEPT_

// Headers //---------------------------------------------------------------------------------------
#include <gnx_list.hpp>

// Concepts Declaration //--------------------------------------------------------------------------

//---------------------------------------------------------------------------------------gnx_concept
/* Associates a concept with an index. */
template <int ID,class OBS = gnx_nil,class = void> struct gnx_concept : gnx_type<gnx_nil> {};

//------------------------------------------------------------------------------------gnx_nb_concept
/* Returns the total number of indexed concepts. */
template <class OBS,int N,bool TEST> struct _gnx_nb_concept_if_;

template <class OBS,int N = 0> struct gnx_nb_concept
: _gnx_nb_concept_if_<OBS,N,gnx_same<typename gnx_concept<N+1,OBS>::type,gnx_nil>::value> {};

template <class OBS,int N> struct _gnx_nb_concept_if_<OBS,N,true>
: gnx_value<int,N> {};

template <class OBS,int N> struct _gnx_nb_concept_if_<OBS,N,false>
: gnx_nb_concept<OBS,N+1> {};

//------------------------------------------------------------------------------------gnx_concept_id
/* Returns the index of a concept. */
template <class CONCEPT> struct gnx_concept_id
: gnx_value<int,gnx_nb_concept<CONCEPT>::value+1> {};

//---------------------------------------------------------------------------------gnx_index_concept
/* Macro that indexes a concept. */
#ifdef gnx_cpp11
 #define gnx_index_concept(...)                                                             \
  template <class OBS>                                                                      \
  struct gnx_concept<gnx_nb_concept< __VA_ARGS__ >::value+1,OBS,                            \
                     typename gnx_enable_if< gnx_not< gnx_same<OBS, __VA_ARGS__ > > >::type \
                    > : gnx_type< __VA_ARGS__ > {}                                          \

#else
 #define gnx_index_concept(CONCEPT)                                                     \
  template <class OBS>                                                                  \
  struct gnx_concept<gnx_nb_concept< CONCEPT >::value+1,OBS,                            \
                     typename gnx_enable_if< gnx_not< gnx_same<OBS, CONCEPT > > >::type \
                    > : gnx_type< CONCEPT > {}                                          \

#endif

//-------------------------------------------------------------------------------gnx_declare_concept
/* Macro that declares a concept, i.e. declares and indexes the associated type. */
#define gnx_declare_concept(CONCEPT) \
 struct CONCEPT {};                  \
 gnx_index_concept(CONCEPT)          \

//----------------------------------------------------------------------------gnx_declare_ns_concept
/* Macro that declares a concept inside a namespace. */
#define gnx_declare_ns_concept(NAMESPACE,CONCEPT) \
 namespace NAMESPACE { struct CONCEPT {}; }       \
 gnx_index_concept(NAMESPACE::CONCEPT)            \

// Concepts Relationships //------------------------------------------------------------------------

//--------------------------------------------------------------------------------gnx_models_concept
/* Associates a type (or a concept) with a concept. */
template <class TYPE_OR_CONCEPT,class CONCEPT> struct gnx_models_concept : gnx_false {};

//-------------------------------------------------------------------------------gnx_direct_concepts
/* Returns a list of all the concepts that a type models directly. */
template <class TYPE,class OBS,class LIST,int N,bool TEST> struct _gnx_direct_concepts_if_1_;
template <class TYPE,class OBS,class LIST,int N,bool TEST> struct _gnx_direct_concepts_if_2_;

template <class TYPE,class OBS = gnx_nil,class LIST = gnx_nil,int N = 1> struct gnx_direct_concepts
: _gnx_direct_concepts_if_1_<TYPE,OBS,LIST,N,
                             gnx_same<typename gnx_concept<N,OBS>::type,gnx_nil>::value
                            > {};

template <class TYPE,class OBS,class LIST,int N>
struct _gnx_direct_concepts_if_1_<TYPE,OBS,LIST,N,true> : gnx_type<LIST> {};

template <class TYPE,class OBS,class LIST,int N>
struct _gnx_direct_concepts_if_1_<TYPE,OBS,LIST,N,false>
: _gnx_direct_concepts_if_2_<TYPE,OBS,LIST,N,
                             gnx_models_concept<TYPE,typename gnx_concept<N,OBS>::type>::value
                            > {};

template <class TYPE,class OBS,class LIST,int N>
struct _gnx_direct_concepts_if_2_<TYPE,OBS,LIST,N,true>
: gnx_direct_concepts<TYPE,OBS,gnx_list<typename gnx_concept<N,OBS>::type,LIST>,N+1> {};

template <class TYPE,class OBS,class LIST,int N>
struct _gnx_direct_concepts_if_2_<TYPE,OBS,LIST,N,false>
: gnx_direct_concepts<TYPE,OBS,LIST,N+1> {};

//--------------------------------------------------------------------------------gnx_merge_concepts
/* Merges two lists of concepts, avoiding duplicates. */
template <class LIST1,class LIST2,bool TEST> struct _gnx_merge_concepts_if_1_;
template <class LIST1,class LIST2,bool TEST> struct _gnx_merge_concepts_if_2_;

template <class LIST1,class LIST2> struct gnx_merge_concepts
: _gnx_merge_concepts_if_1_<LIST1,LIST2,gnx_equals< gnx_concept_id<typename LIST1::content>,
                                                    gnx_concept_id<typename LIST2::content>
                                                  >::value
                           > {};

template <class LIST1,class LIST2> struct _gnx_merge_concepts_if_1_<LIST1,LIST2,true>
: gnx_type< gnx_list< typename LIST1::content,
                      typename gnx_merge_concepts< typename LIST1::next,
                                                   typename LIST2::next
                                                 >::type
                    >
          > {};

template <class LIST1,class LIST2> struct _gnx_merge_concepts_if_1_<LIST1,LIST2,false>
: _gnx_merge_concepts_if_2_<LIST1,LIST2,
                            gnx_greater< gnx_concept_id<typename LIST1::content>,
                                         gnx_concept_id<typename LIST2::content>
                                       >::value
                           > {};

template <class LIST1,class LIST2> struct _gnx_merge_concepts_if_2_<LIST1,LIST2,true>
: gnx_type< gnx_list< typename LIST1::content,
                      typename gnx_merge_concepts< typename LIST1::next,
                                                   LIST2
                                                 >::type
                    >
          > {};

template <class LIST1,class LIST2> struct _gnx_merge_concepts_if_2_<LIST1,LIST2,false>
: gnx_type< gnx_list< typename LIST2::content,
                      typename gnx_merge_concepts< LIST1,
                                                   typename LIST2::next
                                                 >::type
                    >
          > {};

template <class LIST1>
struct gnx_merge_concepts<LIST1,gnx_nil>
: gnx_type< gnx_list< typename LIST1::content,
                      typename gnx_merge_concepts< typename LIST1::next,
                                                   gnx_nil
                                                 >::type
                    >
          > {};

template <class LIST2>
struct gnx_merge_concepts<gnx_nil,LIST2>
: gnx_type< gnx_list< typename LIST2::content,
                      typename gnx_merge_concepts< gnx_nil,
                                                   typename LIST2::next
                                                 >::type
                    >
          > {};

template <> struct gnx_merge_concepts<gnx_nil,gnx_nil> : gnx_type<gnx_nil> {};

//----------------------------------------------------------------------------------gnx_all_concepts
/* Returns a list of all the concepts that a type models directly or indirectly. */
template <class TYPE,class OBS = gnx_nil> struct gnx_all_concepts;

template <class CONCEPTS,class OBS> struct _gnx_search_all_concepts_
: gnx_merge_concepts< typename gnx_all_concepts<typename CONCEPTS::content,OBS>::type,
                      typename _gnx_search_all_concepts_<typename CONCEPTS::next,OBS>::type
                    > {};

template <class OBS> struct _gnx_search_all_concepts_<gnx_nil,OBS> : gnx_type<gnx_nil> {};

template <class TYPE,class OBS> struct gnx_all_concepts
: gnx_merge_concepts< typename gnx_direct_concepts<TYPE,OBS>::type,
                      typename _gnx_search_all_concepts_
                               <typename gnx_direct_concepts<TYPE,OBS>::type,OBS>::type
                    > {};

//-------------------------------------------------------------------------------gnx_matches_concept
/* Returns whether a type (or a concept) models (or refines) a concept directly or indirectly. */
template <class TYPE,class CONCEPT,class OBS = gnx_nil> struct gnx_matches_concept
: gnx_search_list<CONCEPT,typename gnx_all_concepts<TYPE,OBS>::type> {};

template <class TYPE,class OBS> struct gnx_matches_concept<TYPE,TYPE,OBS> : gnx_true {};
template <class TYPE,class OBS> struct gnx_matches_concept<TYPE,gnx_nil,OBS> : gnx_true {};

// Logical Operations //----------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------gnx_cor
/* Represents the logical "or" operation between two concepts. */
template <class CONCEPT1,class CONCEPT2> struct gnx_cor {};

// Ordering: X and Y refine cor<X,Y> //
template <class CONCEPT1,class CONCEPT2>
struct gnx_models_concept< CONCEPT1,gnx_cor<CONCEPT1,CONCEPT2> > : gnx_true {};

template <class CONCEPT1,class CONCEPT2>
struct gnx_models_concept< CONCEPT2,gnx_cor<CONCEPT1,CONCEPT2> > : gnx_true {};

//------------------------------------------------------------------------------------------gnx_cand
/* Represents the logical "and" operation between two concepts. */
template <class CONCEPT1,class CONCEPT2> struct gnx_cand {};

// Matching Specialization //
template <class TYPE,class CONCEPT1,class CONCEPT2,class OBS>
struct gnx_matches_concept<TYPE,gnx_cand<CONCEPT1,CONCEPT2>,OBS>
: gnx_and< gnx_matches_concept<TYPE,CONCEPT1,OBS>,
           gnx_matches_concept<TYPE,CONCEPT2,OBS>
         > {};

// Ordering: cand<X,Y> refines X and Y //
template <class CONCEPT1,class CONCEPT2>
struct gnx_models_concept<gnx_cand<CONCEPT1,CONCEPT2>,CONCEPT1> : gnx_true {};

template <class CONCEPT1,class CONCEPT2>
struct gnx_models_concept<gnx_cand<CONCEPT1,CONCEPT2>,CONCEPT2> : gnx_true {};

//------------------------------------------------------------------------------------------gnx_cnot
/* Represents the logical "not" operation applied on a concept. */
template <class CONCEPT> struct gnx_cnot {};

// Matching Specialization //
template <class TYPE,class CONCEPT,class OBS>
struct gnx_matches_concept<TYPE,gnx_cnot<CONCEPT>,OBS>
: gnx_not< gnx_matches_concept<TYPE,CONCEPT,OBS> > {};

//---------------------------------------------------------------------------------------_gnx_comma_
/* Macro that represents a comma, useful to insert a comma in a macro parameter. */
#define _gnx_comma_ ,

//------------------------------------------------------------------------------gnx_index_or_concept
/* Macro that indexes a logical "or" operation between two concepts. */
#define gnx_index_or_concept(CONCEPT1,CONCEPT2)            \
 gnx_index_concept(gnx_cor<CONCEPT1 _gnx_comma_ CONCEPT2>) \

//-----------------------------------------------------------------------------gnx_index_and_concept
/* Macro that indexes a logical "and" operation between two concepts. */
#define gnx_index_and_concept(CONCEPT1,CONCEPT2)            \
 gnx_index_concept(gnx_cand<CONCEPT1 _gnx_comma_ CONCEPT2>) \

//-----------------------------------------------------------------------------gnx_index_not_concept
/* Macro that indexes a logical "not" operation applied on a concept. */
#define gnx_index_not_concept(CONCEPT) \
 gnx_index_concept(gnx_cnot<CONCEPT>)  \

//-----------------------------------------------------------------------------------gnx_op_concepts
/* Returns a list of all the logical operations that a type models directly or indirectly. */
template <class TYPE,class OBS = gnx_nil,int N = 1,
          class CANDIDATE = typename gnx_concept<N,OBS>::type>
struct gnx_op_concepts : gnx_op_concepts<TYPE,OBS,N+1> {};

template <class TYPE,int N,class OBS>
struct gnx_op_concepts<TYPE,OBS,N,gnx_nil> : gnx_type<gnx_nil> {};

template <class TYPE,class OBS,int N,class CANDIDATE,bool TEST>
struct _gnx_check_op_concept_if_;

template <class TYPE,class OBS,int N,class CANDIDATE>
struct _gnx_check_op_concept_
: _gnx_check_op_concept_if_<TYPE,OBS,N,CANDIDATE,
                            gnx_matches_concept<TYPE,CANDIDATE,OBS>::value
                           > {};

template <class TYPE,class OBS,int N,class CANDIDATE>
struct _gnx_check_op_concept_if_<TYPE,OBS,N,CANDIDATE,true>
: gnx_type< gnx_list< CANDIDATE,
                      typename gnx_op_concepts<TYPE,OBS,N+1>::type
                    >
          > {};

template <class TYPE,class OBS,int N,class CANDIDATE>
struct _gnx_check_op_concept_if_<TYPE,OBS,N,CANDIDATE,false>
: gnx_op_concepts<TYPE,OBS,N+1> {};

template <class TYPE,class OBS,int N,class CONCEPT1,class CONCEPT2>
struct gnx_op_concepts< TYPE,OBS,N,gnx_cor<CONCEPT1,CONCEPT2> >
: _gnx_check_op_concept_< TYPE,OBS,N,gnx_cor<CONCEPT1,CONCEPT2> > {};

template <class TYPE,class OBS,int N,class CONCEPT1,class CONCEPT2>
struct gnx_op_concepts< TYPE,OBS,N,gnx_cand<CONCEPT1,CONCEPT2> >
: _gnx_check_op_concept_< TYPE,OBS,N,gnx_cand<CONCEPT1,CONCEPT2> > {};

template <class TYPE,class OBS,int N,class CONCEPT>
struct gnx_op_concepts< TYPE,OBS,N,gnx_cnot<CONCEPT> >
: _gnx_check_op_concept_< TYPE,OBS,N,gnx_cnot<CONCEPT> > {};

// Best Concept Selection //------------------------------------------------------------------------

//----------------------------------------------------------------------------------gnx_uses_concept
/* Associates a concept with a context. */
template <class CONTEXT,class CONCEPT> struct gnx_uses_concept : gnx_false {};

//----------------------------------------------------------------------------gnx_contextual_concept
/* Returns the best concept of a type for a given context. */
template <class CONTEXT,class CONCEPTS,class CANDIDATE,bool TEST> struct _gnx_search_concept_if_;

template <class CONTEXT,class CONCEPTS,class CANDIDATE = gnx_nil> struct _gnx_search_concept_
: _gnx_search_concept_if_<CONTEXT,CONCEPTS,CANDIDATE,
                          gnx_and< gnx_uses_concept<CONTEXT,typename CONCEPTS::content>,
                                   gnx_matches_concept<typename CONCEPTS::content,CANDIDATE,CONTEXT>
                                 >::value
                         > {};

template <class CONTEXT,class CONCEPTS,class CANDIDATE>
struct _gnx_search_concept_if_<CONTEXT,CONCEPTS,CANDIDATE,true>
: _gnx_search_concept_<CONTEXT,typename CONCEPTS::next,typename CONCEPTS::content> {};

template <class CONTEXT,class CONCEPTS,class CANDIDATE>
struct _gnx_search_concept_if_<CONTEXT,CONCEPTS,CANDIDATE,false>
: _gnx_search_concept_<CONTEXT,typename CONCEPTS::next,CANDIDATE> {};

template <class CONTEXT,class CANDIDATE>
struct _gnx_search_concept_<CONTEXT,gnx_nil,CANDIDATE> : gnx_type<CANDIDATE> {};

template <class CONTEXT,class TYPE> struct gnx_contextual_concept
: _gnx_search_concept_< CONTEXT,
                        typename gnx_merge_concepts< typename gnx_all_concepts<TYPE,CONTEXT>::type,
                                                     typename gnx_op_concepts<TYPE,CONTEXT>::type
                                                   >::type
                      > {};

template <class CONTEXT> struct gnx_contextual_concept<CONTEXT,gnx_nil> : gnx_type<gnx_nil> {};
template <class TYPE> struct gnx_contextual_concept<gnx_nil,TYPE> : gnx_type<gnx_nil> {};

//----------------------------------------------------------------------------------gnx_best_concept
/* Macro that returns the best concept of a type for a given context. Obsolete in C++11. */
#define gnx_best_concept(CONTEXT,TYPE)                                               \
 typename gnx_contextual_concept< CONTEXT,typename gnx_base_type<TYPE>::type >::type \

//--------------------------------------------------------------------------------gnx_best_concept_t
/* Helper function that returns the best concept of a type for a given context. C++11 only. */
#ifdef gnx_cpp11
 template <class CONTEXT,class TYPE>
 using gnx_best_concept_t = typename gnx_contextual_concept< CONTEXT,gnx_base_type_t<TYPE> >::type;
#endif

//---------------------------------------------------------------------------------------gnx_matches
/* Returns whether a type (or a concept) matches a concept or is the same as another type.
   The C++11 version allows comparing a type (or a concept) with several other types and concepts,
   and returns whether it matches at least one of them. The first parameter is an observer
   that is mandatory to avoid Schrodinger's cat effect.
*/
#ifdef gnx_cpp11
 template <class OBS,class TYPE1,class TYPE2,class... REST>
 struct gnx_matches : gnx_or< gnx_matches<OBS,TYPE1,TYPE2>,
                              gnx_matches<OBS,TYPE1,REST...>
                            > {};

 template <class OBS,class TYPE1,class TYPE2>
 struct gnx_matches<OBS,TYPE1,TYPE2>
 : gnx_or< gnx_same<TYPE1,TYPE2>,
           gnx_matches_concept<TYPE1,TYPE2,OBS>
         > {};
#else
 template <class OBS,class TYPE1,class TYPE2>
 struct gnx_matches : gnx_or< gnx_same<TYPE1,TYPE2>,
                              gnx_matches_concept<TYPE1,TYPE2,OBS>
                            > {};
#endif

//------------------------------------------------------------------------------------gnx_add_models
/* Macro that adds a modeling (or refinement) relationship between a type (or a concept)
   and a concept. C++11 only. */
#ifdef gnx_cpp11
 #define gnx_add_models(...)                             \
  struct gnx_models_concept< __VA_ARGS__ > : gnx_true {} \

#endif

//--------------------------------------------------------------------------------------gnx_add_uses
/* Macro that associates a concept with a context. C++11 only. */
#ifdef gnx_cpp11
 #define gnx_add_uses(...)                             \
  struct gnx_uses_concept< __VA_ARGS__ > : gnx_true {} \

#endif

//------------------------------------------------------------------------------------gnx_freeze_obs
/* Macro that explicitly "freezes" concepts for an observer. It allows extending concepts
   after a full instantiation of "gnx_matches_concept" without a "partial specialization after
   instantiation" error from recent compilers.
*/
#ifdef gnx_cpp11
 #define gnx_freeze_obs(...) gnx_index_concept(__VA_ARGS__)
#else
 #define gnx_freeze_obs(OBSERVER) gnx_index_concept(OBSERVER)
#endif

// End //-------------------------------------------------------------------------------------------
#endif
